# Copyright (C) 2019  Collin Capano
# This program is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 3 of the License, or (at your
# option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
# Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

"""Utilities for handling chain data."""

from __future__ import absolute_import

import numpy

from epsie import array2dict


class ChainData(object):
    """Handles reading and adding data to a chain.

    When initialized, a list of parameter names must be provided for which the
    data will be stored. Items may then be added by giving a dictionary mapping
    parameter names to their values. See the Examples below. If the index given
    where the new data should be added is larger then the length of the
    instance's current memory, it will automatically be extended by the amount
    needed. Scratch space may be allocated ahead of time by using the
    ``extend`` method.

    Data can be retrieved using the ``.data`` attribute, which will return the
    data as a numpy structred array.

    Space for multiple temperatures may be specified by providing the
    ``ntemps`` argument. In this case, the array will have shape
    ``niterations x ntemps``.

    .. note::

        Note that the temperatures are the last index. This is because numpy is
        row major. When stepping a chain, the collection of temperatures at a
        given iteration are often accessed, to write data, and to do
        temperature swaps. However, once the chain is complete, it is more
        common to access all the iterations at once for a given temperature;
        e.g., to calculate autocorrelation length. For this reason, it is
        recommended that the chain data be transposed before doing other
        operations and writing to disk.

    Parameters
    ----------
    parameters : list or tuple of str
        The names of the parameters to store data for.
    dtypes : dict, optional
        Dictionary mapping parameter names to types. Will default to using
        ``float`` for any parameters that are not provided.
    ntemps : int, optional
        The number of temperatures used by the chain. Default is 1.

    Attributes
    ----------
    parameters : tuple
        The names of the parameters that were given.
    dtypes : dict
        The data type used for each of the parameters.
    data

    Examples
    --------
    Create an scratch space for two parameters, "x" and "y". Note that,
    initially, the data attribute returns None, and the length is zero:

    >>> from epsie.chain import ChainData
    >>> chaindata = ChainData(['x', 'y'])
    >>> print(len(chaindata))
    0
    >>> print(chaindata.data)
    None

    Now add some data by passing a dictionary of values. Note that the length
    is automatically extended to accomodate the given index, with zeroes filled
    in up to that point:

    >>> chaindata[1] = {'x': 2.5, 'y': 1.}
    >>> chaindata.data
    array([(0. , 0.), (2.5, 1.)], dtype=[('x', '<f8'), ('y', '<f8')])
    >>> len(chaindata)
    2

    Manually extend the scratch space, and fill it. Note that we can set
    multiple values at once using standard slicing syntax:

    >>> chaindata.extend(4)
    >>> chaindata[2:] = {'x': [3.5, 4.5, 5.5, 6.5], 'y': [2, 3, 4, 5]}
    >>> chaindata.data
    array([(0. , 0.), (2.5, 1.), (3.5, 2.), (4.5, 3.), (5.5, 4.), (6.5, 5.)],
      dtype=[('x', '<f8'), ('y', '<f8')])

    Since we did not specify dtypes, the data types have all defaulted to
    floats. Change 'y' to be ints instead:

    >>> chaindata.dtypes = {'y': int}
    >>> chaindata.data
    array([(0. , 0), (2.5, 1), (3.5, 2), (4.5, 3), (5.5, 4), (6.5, 5)],
      dtype=[('x', '<f8'), ('y', '<i8')])

    Clear the memory, and set the new length to be 3:

    >>> chaindata.clear(3)
    >>> chaindata.data
    array([(0., 0), (0., 0), (0., 0)], dtype=[('x', '<f8'), ('y', '<i8')])

    """

    def __init__(self, parameters, dtypes=None, ntemps=1):
        self.parameters = tuple(parameters)
        self._data = None
        self._dtypes = {}
        if dtypes is None:
            dtypes = {}
        self.dtypes = dtypes  # will call the dtypes.setter, below
        self.ntemps = ntemps

    @property
    def data(self):
        """Returns the saved data as a numpy structered array.

        If no data has been added yet, and an initial length was not specified,
        returns ``None``.
        """
        try:
            return self._data
        except AttributeError as e:
            if self._data is None:
                return None
            else:
                raise AttributeError(e)

    @property
    def dtypes(self):
        """Dictionary mapping parameter names to their data types."""
        return self._dtypes

    @dtypes.setter
    def dtypes(self, dtypes):
        """Sets/updates the dtypes to the given dictionary.

        If data has already been saved for a parameter, an attempt will be
        made to cast the data to the new data type.

        A ``ValueError`` will be raised if a parameter name is in the
        dictionary that was not provided upon initialization.

        Parameters
        ----------
        dtypes : dict
            Dictionary mapping parameter names to data types. The data type
            of any parameters not provided will remain their current/default
            (float) type.
        """
        unrecognized = [p for p in dtypes if p not in self.parameters]
        if any(unrecognized):
            raise ValueError("unrecognized parameter(s) {}"
                             .format(', '.join(unrecognized)))
        # store it
        self._dtypes.update(dtypes)
        # fill in any missing parameters
        self._dtypes.update({p: float for p in self.parameters
                             if p not in self._dtypes})
        # create the numpy verion
        self._npdtype = numpy.dtype([(p, self.dtypes[p])
                                     for p in self.parameters])
        # cast to new dtype if data already exists
        if self._data is not None:
            self._data = self._data.astype(self._npdtype)

    def __len__(self):
        if self._data is None:
            return 0
        else:
            return self._data.shape[0]

    def extend(self, n):
        """Extends scratch space by n items.
        """
        if self.ntemps == 1:
            newshape = n
        else:
            newshape = (n, self.ntemps)
        new = numpy.zeros(newshape, dtype=self._npdtype)

        if self._data is None:
            self._data = new
        else:
            self._data = numpy.append(self._data, new, axis=0)

    def set_len(self, n):
        """Sets the data length to ``n``.

        If the data length is already > ``n``, a ``ValueError`` is raised.
        """
        lself = len(self)
        if lself < n:
            self.extend(n - lself)
        else:
            raise ValueError("current length ({}) is already greater than {}"
                             .format(lself, n))

    def clear(self, newlen=None):
        """Clears the memory.

        Parameters
        ----------
        newlen : int, optional
            If provided, will create new scratch space with the given length.
        """
        self._data = None
        if newlen is None:
            newlen = 0
        self.extend(newlen)

    def __repr__(self):
        return repr(self.data)

    def __getitem__(self, index):
        if self._data is None:
            raise ValueError("no data has been set yet")
        return self._data[index]

    def asdict(self, index=None):
        """Returns the data as a dictionary.

        Parameters
        ----------
        index : slice, optional
            Only get the elements indicated by the given slice before
            converting to a dictionary.
        """
        if index is None:
            return array2dict(self.data)
        else:
            return array2dict(self[index])

    def __setitem__(self, index, value):
        # try to get the element to set; if it fails, then try extending the
        # data by the amount needed
        try:
            elem = self._data[index]
        except (IndexError, TypeError):  # get TypeError if _data is None
            self.extend(index + 1 - len(self))
            elem = self._data[index]
        # if value is a dictionary and index is not a string, then we're
        # setting values in the array by dictionary
        if isinstance(value, dict) and not isinstance(index, str):
            for p in value:
                elem[p] = value[p]
        # otherwise, just fall back to using the structred array's setitem
        else:
            self._data[index] = value


def detect_dtypes(data):
    """Convenience function to detect the dtype of some data.

    Parameters
    ----------
    data : dict or numpy.ndarray
        Either a numpy structred array/void or a dictionary mapping parameter
        names to some (arbitrary) values. The values may be either arrays or
        atomic data. If the former, the dtype will be taken from the array's
        dtype.

    Returns
    -------
    dict :
        Dictionary mapping the parameter names to types.
    """
    if not isinstance(data, dict):  # assume it's a numpy.void or numpy.ndarray
        data = array2dict(data)
    return {p: val.dtype if isinstance(val, numpy.ndarray) else type(val)
            for (p, val) in data.items()}



# Copyright (C) 2019  Collin Capano
# This program is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 3 of the License, or (at your
# option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
# Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

from __future__ import absolute_import

import numpy
from scipy import stats

from .base import BaseProposal


class Normal(BaseProposal):
    """Uses a normal distribution with a fixed variance for proposals.

    This proposal may handle one or more parameters.

    Parameters
    ----------
    parameters : (list of) str
        The names of the parameters to produce proposals for.
    cov : array, optional
        The covariance matrix of the parameters. May provide either a single
        float, a 1D array with length ``ndim``, or an ``ndim x ndim`` array,
        where ``ndim`` = the number of parameters given. If a single float or
        a 1D array is given, will use a diagonal covariance matrix (i.e., all
        parameters are independent of each other). Default (None) is to use
        unit variance for all parameters.

    Attributes
    ----------
    cov : numpy.ndarray
        The covariance matrix being used.
    """

    name = 'normal'
    symmetric = True

    def __init__(self, parameters, cov=None):
        self.parameters = parameters
        self.ndim = len(parameters)
        self.isdiagonal = False
        self._cov = None
        self.cov = cov

    @property
    def cov(self):
        """The covariance matrix used.
        """
        return numpy.diag(self._cov)

    @cov.setter
    def cov(self, cov):
        """Sets the covariance matrix.

        If a single float or a 1D array is given, will use a diagonal
        covariance matrix (i.e., all parameters are independent of each other).
        Default (None) is to use unit variance for all parameters.

        Raises a ``ValueError`` if the dimensionality of the given array
        isn't ndim x ndim.
        """
        if cov is None:
            cov = 1.
        if not isinstance(cov, numpy.ndarray):
            cov = numpy.array(cov)
        # make sure cov is atleast 1D array
        if cov.ndim < 1 or cov.size == 1:
            cov = numpy.repeat(cov.item(), len(self.parameters))
        # if its a 1D array, means diagonal covariance
        if cov.ndim == 1:
            self.isdiagonal = True
        else:
            # check that dimensionality makes sense
            if cov.shape != (self.ndim, self.ndim):
                raise ValueError("dimension of covariance matrix does not "
                                 "match given number of parameters")
            # check if off-diagonal terms are all zero, if so, just store
            # the diagonals
            self.isdiagonal = (
                cov[~numpy.eye(cov.shape[0], dtype=bool)] == 0).all()
            if self.isdiagonal:
                cov = cov[numpy.diag_indices(cov.shape[0])]
        self._cov = cov

    @property
    def state(self):
        return {'random_state': self.random_state}

    def set_state(self, state):
        self.random_state = state['random_state']

    def jump(self, fromx):
        # the normal RVS is much faster than the multivariate one, so use it
        # if we can
        if self.isdiagonal:
            mu = [fromx[p] for p in self.parameters]
            newpt = self.random_generator.normal(mu, self._cov)
        else:
            newpt = self.random_generator.multivariate_normal(
                [fromx[p] for p in self.parameters], self.cov)
        return dict(zip(self.parameters, newpt))


    def logpdf(self, xi, givenx):
        means = [givenx[p] for p in self.parameters]
        xi = [xi[p] for p in self.parameters]
        if self.isdiagonal:
            logp = stats.normal.logpdf(xi, loc=means, scale=self._cov)
        else:
            logp = stats.multivariate_normal(xi, mean=means, cov=self._cov)
        return logp


class AdaptiveNormal(Normal):
    r"""Uses a Gaussian distribution with adaptive variance.

    The adaptation algorithm is based on Eqs. 35 and 36 of [1]_.
    The size of the variance at each step is based on the width of the prior
    and whether or not the previous proposal was accepted or not. See Notes
    for more details.

    Multiple parameters may be given. However, currently, only independent
    variables are supported (i.e., the covariance matrix must be diagonal).

    Parameters
    ----------
    parameters : (list of) str
        The names of the parameters.
    prior_widths : dict
        Dictionary mapping parameter names to values giving the width of each
        parameter's prior. The values may be floats, or any object that has
        an ``__abs__`` method that will return a float.
    adaptation_duration : int
        The number of iterations over which to apply the adaptation. No more
        adaptation will be done once a chain exceeds this value.
    adaptation_decay : int, optional
        The decay rate to use for the adaptation size (:math:`beta` in the
        equation below). If not provided, will use :math:`1/\log_10(T)`, where
        :math:`T` is the adaptation duration.
    start_iteration : int, optional
        The iteration to start doing the adaptation (:math:`k_0+1` in the
        equation below). Must be greater than zero. Default is 1.
    target_rate : float, optional
        The target acceptance rate. Default is 0.234.
    initial_var : array, optional
        The initial variance to use. Default is to use
        `(1 - target_rate)*0.09*prior_widths`.

    Notes
    -----
    For a given parameter, the variance of the :math:`k`th iteration is given
    by [1]_:

    .. math::

        \sigma_k = \sigma_{k-1} + \alpha_{k-1}
            \left[\left(k - k_0\right)^{-\beta} - 0.1\right]\frac{\Delta}{10},

    where :math:`\alpha_{k-1} = 1 - \xi` if the previous iteration was
    accpeted and :math:`\alpha_{k-1} = -\xi` if the previous iteration was
    rejected. Here, :math:`\xi` is the target acceptance rate, :math:`\Delta`
    is the prior width, :math:`\beta` is the adaptation decay, and :math:`k_0`
    gives the iteration after which the adaptation begins. The initial variance
    :math:`\sigma_0` to use is a free parameter. The default in this function
    is to use :math:`\sigma_0 = (1-\xi)0.09\Delta`.


    References
    ----------
    .. [1] J. Veitch et al., "Parameter estimation for compact binaries with
        ground-based gravitational-wave observations using the LALInference
        software library, " Phys. Rev. D91 042003 (2015),
        arXiv:1409.7215 [gr-qc].
    """
    name = 'adaptive_normal'
    symmetric = True

    def __init__(self, parameters, prior_widths, adaptation_duration,
                 adaptation_decay=None, start_iteration=1, target_rate=0.234,
                 initial_var=None):
        # set the parameters, initialize the covariance matrix
        super(AdaptiveNormal, self).__init__(parameters)
        # check that a diagonal covariance was provided
        if not self.isdiagonal:
            raise ValueError("Only independent variables are supported "
                             "(all off-diagonal terms in the covariance "
                             "must be zero)")
        # figure out initial variance to use
        self._deltas = None
        self.prior_widths = prior_widths
        self._adaptation_duration = None
        self.adaptation_duration = adaptation_duration
        if adaptation_decay is None:
            adaptation_decay = 1./numpy.log10(self.adaptation_duration)
        self.adaptation_decay = adaptation_decay
        self._start_iteration = None
        self.start_iteration = start_iteration
        self.target_rate = target_rate
        if initial_var is None:
            initial_var = numpy.diag((1 - self.target_rate)*0.09*self.deltas)
        # set the covariance to the initial
        self.cov = initial_var

    @property
    def prior_widths(self):
        """The width of the prior used for each parameter."""
        return self._prior_widths

    @prior_widths.setter
    def prior_widths(self, prior_widths):
        """Sets the prior widths, making sure that widths are provided for
        each parameter in ``parameters``.

        Also sets the deltas attribute.
        """
        try:
            self._prior_widths = {p: abs(prior_widths[p])
                                  for p in self.parameters}
        except KeyError:
            raise ValueError("must provide prior widths for every parameter")
        self._deltas = numpy.array([self.prior_widths[p]
                                    for p in self.parameters])

    @property
    def deltas(self):
        """The prior widths, as a numpy array."""
        return self._deltas

    @property
    def start_iteration(self):
        """The iteration that the adaption begins."""
        return self._start_iteration

    @start_iteration.setter
    def start_iteration(self, start_iteration):
        """Sets the start iteration, making sure it is >= 1."""
        if start_iteration < 1:
            raise ValueError("start_iteration must be >= 1")
        self._start_iteration = start_iteration

    @property
    def adaptation_duration(self):
        """The adaptation duration used."""
        return self._adaptation_duration

    @adaptation_duration.setter
    def adaptation_duration(self, adaptation_duration):
        """Sets the adaptation duration to the given value, making sure it is
        larger than 1.
        """
        if adaptation_duration < 1:
            raise ValueError("adaptation duration must be >= 1")
        self._adaptation_duration = adaptation_duration

    def update(self, chain):
        """Updates the adaptation based on whether the last jump was accepted.

        This prepares the proposal for the next jump.
        """
        # subtact 1 from the start iteration, since the update happens after
        # the jump
        dk = chain.iteration - (self.start_iteration - 1)
        if 1 <= dk < self.adaptation_duration:
            dk = dk**(-self.adaptation_decay) - 0.1
            if chain.acceptance[-1]['accepted']:
                alpha = 1 - self.target_rate
            else:
                alpha = -self.target_rate
            dsigmas = alpha * dk * self.deltas/10.
            # ensure we don't go negative
            newcov = self._cov + dsigmas
            lzidx = newcov < 0
            newcov[lzidx] = self._cov[lzidx]
            self._cov = newcov

    @property
    def state(self):
        return {'random_state': self.random_state,
                'cov': self._cov}

    def set_state(self, state):
        self.random_state = state['random_state']
        self._cov = state['cov']



# Copyright (C) 2019  Collin Capano
# This program is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 3 of the License, or (at your
# option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
# Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

"""Base class and functions for samplers."""

from __future__ import absolute_import

import itertools
from abc import (ABCMeta, abstractmethod)
from six import add_metaclass
import numpy

from epsie import (create_seed, dump_state, load_state)
from epsie.proposals import Normal


@add_metaclass(ABCMeta)
class BaseSampler(object):
    """Base class for samplers.
    """
    _parameters = None
    _proposals = None
    _model = None
    _chains = None
    _seed = None
    _map = None

    @property
    def parameters(self):
        """The sampled parameters."""
        return self._parameters

    @parameters.setter
    def parameters(self, parameters):
        if not isinstance(parameters, (list, tuple, numpy.ndarray)):
            parameters = [parameters]
        self._parameters = tuple(sorted(parameters))

    @property
    def proposals(self):
        """Dictionary of the proposals used for the sampled parameters."""
        return self._proposals

    def set_proposals(self, proposals=None, default_proposal=None,
                      default_proposal_args=None):
        """Sets the proposals for all parameters.

        If no proposals are given for one or more parameters, the default
        proposal will be used for them.

        Parameters
        ----------
        proposals : dict, optional
            A dictionary of parameter names ->
            :py:class:`proposals.BaseProposal` instances. If none provided,
            will use the ``default_proposal`` for all parameters.
        default_proposal : proposals.BaseProposal type class, optional
            The default proposal class to use for parameters that are not
            specified in ``proposals``. Default (None) is to use
            :py:class:`proposals.Normal`.
        default_proposal_args : dict, optional
            Dictionary of keyword arguments to pass to the default proposal
            when initializing.
        """
        if proposals is None:
            proposals = {}
        # create default proposal instances for the other parameters
        if default_proposal is None:
            default_proposal = Normal
        if default_proposal_args is None:
            default_proposal_args = {}
        missing_props = tuple(set(self.parameters)
                              - set(itertools.chain(*proposals.keys())))
        if missing_props:
            proposals[missing_props] = default_proposal(
                missing_props, **default_proposal_args)
        self._proposals = proposals

    @property
    def model(self):
        """The model used."""
        return self._model

    @model.setter
    def model(self, model):
        self._model = model

    @property
    def chains(self):
        """List of the chains used."""
        return self._chains

    @chains.setter
    def chains(self, chains):
        self._chains = chains

    @property
    def seed(self):
        """The seed used for the random bit generator."""
        return self._seed

    @seed.setter
    def seed(self, seed):
        """Sets the seed. If the given seed is None, will create one."""
        if seed is None:
            seed = create_seed(seed)
        self._seed = seed

    @property
    def map(self):
        """The function used for parallelizing chain evolution."""
        return self._map

    def set_map(self, pool=None):
        """Sets the map function using the given pool.

        Parameters
        ----------
        pool : Pool object, optional
            The pool of processes to use for parallelization. Must have a
            map function. If none provided, will use the builtin ``map`` (i.e.,
            no parallelization will be used).
        """
        if pool is None:
            self._map = map
        else:
            self._map = pool.map

    @property
    def nchains(self):
        """The number of chains being used."""
        return len(self.chains)

    @property
    def niterations(self):
        """The number of iterations the chains have been run for.
        """
        # all of the chains should be at the same iteration, so just use the
        # first one
        return self.chains[0].iteration

    @abstractmethod
    def _concatenate_dicts(self, attr):
        """Concatenates dictionary attributes over all of the chains.

        This is a convenience function used by properties such as
        ``current_positions`` to gather all of the dictionary attributes from
        the chains.

        Parameters
        ----------
        attr : str
            The name of the attribute to get from the chains. The attribute
            is assumed to return a dictionary.

        Returns
        -------
        dict :
            Dictionary mapping parameters to arrays.
        """
        pass

    @abstractmethod
    def _concatenate_arrays(self, attr, item=None):
        """Concatenates the given attribute over all of the chains.

        This is a convenience function used by properties such as ``positions``
        to gather all of the array attributes from the chains.

        Parameters
        ----------
        attr : str
            The name of the attribute to get from the chains. The attribute
            is assumed to return a (structred) array.
        item : str or array index, optional
            Get a particular item from the (structred) array from each chain
            before concatenating.

        Returns
        -------
        array :
            The returned array sould have ``nchains x niterations`` as the last
            two dimensions.
        """
        pass

    @property
    def start_position(self):
        """The start position of the chains.

        Will raise a ``ValueError`` is ``set_start`` hasn't been run.

        Returns
        -------
        dict :
            Dictionary mapping parameters to arrays with shape ``nchains``
            giving the starting position of each chain.
        """
        return self._concatenate_dicts('start_position')

    @start_position.setter
    def start_position(self, positions):
        """Sets the starting position of all of the chains.

        Parameters
        ----------
        positions : dict
            Dictionary mapping parameter names to arrays of values. The arrays
            must have shape ``[ntemps x] nchains``.
        """
        for (ii, chain) in enumerate(self.chains):
            chain.start_position = {p: positions[p][..., ii]
                                    for p in positions}

    def run(self, niterations):
        """Evolves all of the chains by niterations.

        Parameters
        ----------
        niterations : int
            The number of iterations to evolve the chains for.
        """
        # extend the scratch space for the chains
        for c in self.chains:
            c.scratchlen += max(niterations - (c.scratchlen - len(c)), 0)
        # construct arguments for passing to the pool
        args = zip([niterations]*len(self.chains), self.chains)
        # pylint: disable=locally-disabled, not-callable
        self.chains = list(self.map(_evolve_chain, args))

    def clear(self):
        """Clears all of the chains."""
        for chain in self.chains:
            chain.clear()

    @property
    def lastclear(self):
        """The iteration that the last clear occurred at."""
        # all of the chains should of been cleared at the same time, so just
        # use the first one
        return self.chains[0].lastclear

    @property
    def positions(self):
        """The history of positions from all of the chains."""
        return self._concatenate_arrays('positions')

    @property
    def current_positions(self):
        """The current position of the chains.

        This will default to the start position if the chains haven't been
        run yet.
        """
        return self._concatenate_dicts('current_position')

    @property
    def stats(self):
        """The history of stats from all of the chains."""
        return self._concatenate_arrays('stats')

    @property
    def current_stats(self):
        """The current stats of the chains.

        This will default to the stats of the start positions if the chains
        haven't been run yet.
        """
        return self._concatenate_dicts('current_stats')

    @property
    def blobs(self):
        """The history of all of the blobs from all of the chains."""
        if self.chains[0].hasblobs:
            blobs = self._concatenate_arrays('blobs')
        else:
            blobs = None
        return blobs

    @property
    def current_blobs(self):
        """The current blob data.

        This will default to the blob data of the start positions if the
        chains haven't been run yet.
        """
        if self.chains[0].hasblobs:
            blobs = self._concatenate_dicts('current_blob')
        else:
            blobs = None
        return blobs

    @property
    def acceptance(self):
        """The history of all acceptance stats from all of the chains."""
        return self._concatenate_arrays('acceptance')

    @property
    def state(self):
        """The state of all of the chains.

        Returns a dictionary mapping chain ids to their states.
        """
        return {chain.chain_id: chain.state for chain in self.chains}

    def set_state(self, state):
        """Sets the state of the all of the chains.

        Parameters
        ----------
        state : dict
            Dictionary mapping chain id to the state to set the chain to. See
            :py:func:`chain.Chain.set_state` for details.
        """
        for ii, chain in enumerate(self.chains):
            chain.set_state(state[ii])

    def checkpoint(self, fp, path=None, dsetname='sampler_state'):
        """Save the sampler's state to an HDF file.

        Parameters
        ----------
        fp : :py:class:h5py.File
            Open file handler to an hdf5 file. The file handler must have
            write permission.
        path : str, optional
            What group to write the state to in the hdf file. Default is the
            top-level.
        dsetname : str, optional
            The name of dataset to store the state to. Default is
            "sampler_state".
        """
        dump_state(self.state, fp, path=path, dsetname=dsetname)

    def set_state_from_checkpoint(self, fp, path=None):
        """Loads a state from an HDF file.

        Parameters
        ----------
        fp : :py:class:h5py.File
            Open file handler to an hdf5 file.
        path : str, optional
            What group to store the file to in the hdf file. Default is the
            top-level ('/').
        """
        self.set_state(load_state(fp, path=path))


def _evolve_chain(niterations_chain):
    """Evolves a chain for some number of iterations.

    This is used by ``Sampler.run`` to evolve a collection of chains in a
    parallel environment. This is not a staticmethod of ``Sampler`` because
    such functions need to be picklable when using python's multiprocessing
    pool.

    Parameters
    ----------
    niterations_chain : tuple of int, chain
        The number of iterations to run on the given chain.
    """
    niterations, chain = niterations_chain
    for _ in range(niterations):
        chain.step()
    return chain



# Copyright (C) 2019  Collin Capano
# This program is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 3 of the License, or (at your
# option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
# Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

from __future__ import absolute_import

import itertools

import numpy

from .base import BaseProposal


class JointProposal(BaseProposal):
    """A collection of jump proposals for multiple parameters.

    Parameters
    ----------
    \*proposals :
        The arguments should provide the constituent proposals to use.
    bit_generator : :py:class:`epsie.BIT_GENERATOR` instance or int, optional
        The random bit generator to use, or an integer/None. If the latter, a
        bit generator will be created using
        :py:func:`epsie.create_bit_generator`.
    """
    name = 'joint'

    # Py3XX: change kwargs to explicit random_state=None
    def __init__(self, *proposals, **kwargs):
        bit_generator = kwargs.pop('bit_generator', None)  # Py3XX: delete line
        self.parameters = itertools.chain(*[prop.parameters
                                            for prop in proposals])
        # check that we don't have multiple proposals for the same parameter
        if len(set(self.parameters)) != len(self.parameters):
            # get the repeated parameters
            repeated = [p for p in self.parameters
                        if self.parameters.count(p) > 1]
            raise ValueError("multiple proposals provided for parameter(s) {}"
                             .format(', '.join(repeated)))
        # the joint proposal is symmetric only if all of the constitutent
        # proposals are also
        self._symmetric = all(prop.symmetric for prop in proposals)
        # set the bit generator
        self.bit_generator = bit_generator
        # have all of the proposals use the same random state
        for prop in proposals:
            prop.bit_generator = self.bit_generator
        # store the proposals as a dict of parameters -> proposal; we can do
        # this since the mapping of parameters to proposals is one/many:one
        self.proposals = {prop.parameters: prop for prop in proposals}

    @property
    def symmetric(self):
        return self._symmetric

    def logpdf(self, xi, givenx):
        return sum(p.logpdf(xi, givenx) for p in self.proposals.values())

    def update(self, chain):
        # update each of the proposals
        for prop in self.proposals.values():
            prop.update(chain)

    def jump(self, fromx):
        out = {}
        for prop in self.proposals.values():
            out.update(prop.jump(fromx))
        return out

    @property
    def state(self):
        # get all of the proposals state
        state = {params: prop.state for params, prop in self.proposals.items()}
        # add the global random state
        state['random_state'] = self.random_state
        return state

    def set_state(self, state):
        # set each proposals' state
        for params, prop in self.proposals.items():
            prop.set_state(state[params])
        # set the state of the random number generator
        self.random_state = state['random_state']



# Copyright (C) 2019  Collin Capano
# This program is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 3 of the License, or (at your
# option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
# Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

"""Classes for individual Markov chains."""

from __future__ import absolute_import

import numpy

from epsie.proposals import JointProposal

from .base import BaseChain
from .chaindata import (ChainData, detect_dtypes)


class Chain(BaseChain):
    """A Markov chain.

    The chain requires a ``model`` to  be provided. This can be any object that
    can be called with keyword arguments that map parameter names to values.
    When called, the object must return a tuple of two or three elements. The
    first element must be a float representing the log likelihood at that
    point; the second must be a float representing the log prior at that point.
    Additionally, the model may return a dictionary as a third element in the
    tuple that maps strings to arbitrary data. This additional data will be
    stored as ``blob`` data.

    Parameters
    ----------
    parameters : list or tuple
        List of the parameter names to sample.
    model : object
        Any object that can be called with keyword arguments that map parameter
        names to values. When called, the object must a tuple of ``logl, logp``
        where ``logp`` is the log prior and ``logl`` is the log likelihood at
        the given point. The model may optionally return a dictionary in
        addition that maps strings to any arbitrary data.
    proposals : list of epsie.proposals instances
        List of proposal classes to use for the parameters. There must be one
        and only one proposal for every parameter. A single proposal may cover
        multiple parameters. Proposals must be instances of classes that
        inherit from :py:class:`epsie.proposals.BaseProposal`.
    bit_generator : :py:class:`epsie.BIT_GENERATOR` instance, optional
        Use the given random bit generator for generating random variates. If
        an int or None is provided, a generator will be created instead using
        ``bit_generator`` as a seed.
    chain_id : int, optional
        An interger identifying which chain this is. Default is 0.
    beta : float, optional
        The inverse temperature of the chain. Default is 1.

    Attributes
    ----------
    parameters
    iteration
    lastclear
    scratchlen
    positions
    stats
    acceptance
    blobs
    start_position
    stats0
    blob0
    current_position
    current_stats
    current_blob
    bit_generator
    random_state
    state
    hasblobs
    chain_id : int or None
        Integer identifying the chain.
    """
    def __init__(self, parameters, model, proposals, bit_generator=None,
                 chain_id=0, beta=1.):
        self.parameters = parameters
        self.model = model
        # combine the proposals into a joint proposal
        self.proposal_dist = JointProposal(*proposals,
                                           bit_generator=bit_generator)
        # store the temp
        self.beta = beta
        self.chain_id = chain_id
        self._iteration = 0
        self._lastclear = 0
        self._scratchlen = 0
        self._positions = ChainData(parameters)
        self._stats = ChainData(['logl', 'logp'])
        self._acceptance = ChainData(['acceptance_ratio', 'accepted'],
                                     dtypes={'accepted': bool})
        self._blobs = None
        self._hasblobs = False
        self._start = None
        self._logp0 = None
        self._logl0 = None
        self._blob0 = None

    @property
    def hasblobs(self):
        """Whether the model returns blobs."""
        return self._hasblobs

    @property
    def iteration(self):
        """The number of times the chain has been stepped."""
        return self._iteration

    @property
    def lastclear(self):
        """Returns the iteration of the last time the chain memory was cleared.
        """
        return self._lastclear

    @property
    def scratchlen(self):
        """The length of the scratch space used."""
        return self._scratchlen

    @scratchlen.setter
    def scratchlen(self, n):
        """Set the scratch length to the given value.

        This will immediately increase the scratch length to ``n``. If the
        chain is already longer than ``n``, this will have no immediate impact.
        However, the next time ``clear`` is called, the scratch length will
        be reset to ``n``.

        Parameters
        ----------
        n : int
            The length to set.
        """
        self._scratchlen = n
        try:
            self._positions.set_len(n)
        except ValueError:
            pass
        try:
            self._stats.set_len(n)
        except ValueError:
            pass
        try:
            self._acceptance.set_len(n)
        except ValueError:
            pass
        if self.hasblobs:
            try:
                self._blobs.set_len(n)
            except ValueError:
                pass

    @property
    def start_position(self):
        """Dictionary mapping parameters to their start position.

        If the start position hasn't been set, raises a ``ValueError``.
        """
        if self._start is None:
            raise ValueError("Starting position not set!")
        return self._start

    @start_position.setter
    def start_position(self, position):
        """Sets the starting position.

        This also evaulates the log likelihood and log prior at the starting
        position, as well as determine if the model returns blob data.

        Parameters
        ----------
        position : dict
            Dictionary mapping parameters to values.
        """
        self._start = position.copy()
        # use start position to determine the dtype for positions
        self._positions.dtypes = detect_dtypes(position)
        # evaluate logl, p at this point
        r = self.model(**position)
        try:
            logl, logp, blob = r
            self._hasblobs = True
        except ValueError:
            logl, logp = r
            blob = None
            self._hasblobs = False
        if self._hasblobs:
            if not isinstance(blob, dict):
                raise TypeError("model must return blob data as a dictionary")
            self._blobs = ChainData(blob.keys(), dtypes=detect_dtypes(blob))
        self.stats0 = {'logl': logl, 'logp': logp}
        self.blob0 = blob

    @property
    def stats0(self):
        """Dictionary of the log likelihood and log prior at the start
        position.

        Raises a ``ValueError`` if the start position has not been set yet.
        """
        # check if the start position is set
        _ = self.start_position  # raises a ValueError if not set
        return self._stats0

    @stats0.setter
    def stats0(self, stats):
        """Sets the starting stats.

        Parameters
        ----------
        dict, numpy structred array of len 1, or numpy.void
            Dictionary or numpy array/void object mapping parameter names to
            the starting statistics.
        """
        # check that we're not starting outside of the prior
        if stats['logp'] == -numpy.inf:
            raise ValueError("starting position is outside of the prior!")
        self._stats0 = stats.copy()

    @property
    def blob0(self):
        """The blob data of the starting position, as a dictionary.

        Raises a ``ValueError`` if ``set_start`` has not been run yet.
        """
        # check if the start position is set
        _ = self.start_position  # raises a ValueError if not set
        if self._blob0 is None:
            blob = None
        else:
            blob = self._blob0
        return blob

    @blob0.setter
    def blob0(self, blob):
        """Sets the starting blob.

        Parameters
        ----------
        blob: dict
            Dictionary mapping blob parameters to their values.
        """
        if blob is not None:
            blob = blob.copy()
            # create scratch for blobs
            if self._blobs is None:
                self._blobs = ChainData(blob.keys(),
                                        dtypes=detect_dtypes(blob))
        self._blob0 = blob

    @property
    def positions(self):
        """The history of all of the positions, as a structred array."""
        if self.iteration == 0:
            raise ValueError("No positions as chain hasn't been stepped "
                             "yet; run step() at least once")
        return self._positions[:len(self)]

    @property
    def stats(self):
        """The log likelihoods and log priors of the positions, as a structured
        array.
        """
        if self.iteration == 0:
            raise ValueError("No stats as chain hasn't been stepped "
                             "yet; run step() at least once")
        return self._stats[:len(self)]

    @property
    def acceptance(self):
        """The history of all of acceptance ratios and accepted booleans,
        as a structred array.
        """
        if self.iteration == 0:
            raise ValueError("No acceptance as chain hasn't been stepped "
                             "yet; run step() at least once")
        return self._acceptance[:len(self)]

    @property
    def blobs(self):
        """The history of all of the blob data, as a structured array.

        If the model does not return blobs, this is just ``None``.
        """
        if self.iteration == 0:
            raise ValueError("No blobs as chain hasn't been stepped "
                             "yet; run step() at least once")
        if self.hasblobs:
            blobs = self._blobs[:len(self)]
        else:
            blobs = None
        return blobs

    @property
    def current_position(self):
        """Dictionary of the current position of the chain."""
        if len(self) == 0:
            pos = self.start_position
        else:
            pos = self._positions.asdict(len(self)-1)
        return pos

    @property
    def current_stats(self):
        """Dictionary giving the log likelihood and log prior of the current
        position.
        """
        if len(self) == 0:
            stats = self.stats0
        else:
            stats = self._stats.asdict(len(self)-1)
        return stats

    @property
    def current_blob(self):
        """Dictionary of the blob data of the current position.

        If the model does not return blobs, just returns ``None``.
        """
        if not self.hasblobs:
            blob = None
        elif len(self) == 0:
            blob = self.blob0
        else:
            blob = self._blobs.asdict(len(self)-1)
        return blob

    def clear(self):
        """Clears memory of the current chain, and sets start position to the
        current position.

        New scratch space will be created with length equal to ``scratch_len``.
        """
        if self._iteration > 0:
            # save position then clear
            self._start = self.current_position
            self._positions.clear(self.scratchlen)
            # save stats then clear
            self._stats0 = self.current_stats
            self._stats.clear(self.scratchlen)
            # clear acceptance info
            self._acceptance.clear(self.scratchlen)
            # save blobs then clear
            if self.hasblobs:
                self._blob0 = self.current_blob
                self._blobs.clear(self.scratchlen)
        self._lastclear = self._iteration
        return self

    def __getitem__(self, index):
        """Returns all of the chain data at the requested index."""
        index = (-1)**(index < 0) * (index % len(self))
        out = {'positions': self._positions[index],
               'stats': self._stats[index],
               'acceptance': self._acceptance[index]
               }
        if self._hasblobs:
            out['blobs'] = self._blobs[index]
        return out

    @property
    def bit_generator(self):
        """The random bit generator being used."""
        return self.proposal_dist.bit_generator

    @property
    def random_generator(self):
        """Returns the random number generator."""
        return self.proposal_dist.random_generator

    @property
    def random_state(self):
        """Returns the current state of the bit generator."""
        return self.proposal_dist.random_state

    @property
    def state(self):
        """Returns the current state of the chain.

        The state consists of everything needed such that setting a chain's
        state using another's state will result in identical results.
        """
        state = {}
        state['chain_id'] = self.chain_id
        state['proposal_dist'] = self.proposal_dist.state
        state['iteration'] = self.iteration
        state['current_position'] = self.current_position
        state['current_stats'] = self.current_stats
        state['hasblobs'] = self.hasblobs
        if self.hasblobs:
            blob = self.current_blob
        else:
            blob = None
        state['current_blob'] = blob
        return state

    def set_state(self, state):
        """Sets the state of the chain using the given dict.

        .. warning::
           Running this will clear the chain's current memory, and replace its
           current position with what is saved in the state.

        Parameters
        ----------
        state : dict
            Dictionary of state values.
        """
        self.chain_id = state['chain_id']
        # set the chain position
        self.clear()
        self._iteration = state['iteration']
        self._lastclear = state['iteration']
        self._start = state['current_position'].copy()
        self.stats0 = state['current_stats']
        self._hasblobs = state['hasblobs']
        self.blob0 = state['current_blob']
        # set the proposals' states
        self.proposal_dist.set_state(state['proposal_dist'])

    def step(self):
        """Evolves the chain by a single step."""
        # get the current position; if this is the first step and set_start
        # hasn't been run, this will raise a ValueError
        current_pos = self.current_position
        current_stats = self.current_stats
        current_blob = self.current_blob
        # create a proposal and test it
        proposal = self.proposal_dist.jump(current_pos)
        r = self.model(**proposal)
        if self._hasblobs:
            logl, logp, blob = r
        else:
            logl, logp = r
            blob = None
        # evaluate
        current_logl = current_stats['logl']
        current_logp = current_stats['logp']
        if logp == -numpy.inf:
            # force a reject
            accept = False
            ar = 0.
        else:
            logar = logp + logl * self.beta \
                    - current_logl * self.beta - current_logp
            if not self.proposal_dist.symmetric:
                logar += self.proposal_dist.logpdf(current_pos, proposal) - \
                         self.proposal_dist.logpdf(proposal, current_pos)
            ar = numpy.exp(logar)
            u = self.random_generator.uniform()
            accept = u <= ar
        if accept:
            pos = proposal
            stats = {'logl': logl, 'logp': logp}
        else:
            # reject
            pos = current_pos
            stats = current_stats
            blob = current_blob
        # save
        index = len(self)
        self._positions[index] = pos
        self._stats[index] = stats
        self._acceptance[index] = (ar, accept)
        if self._hasblobs:
            self._blobs[index] = blob
        self._iteration += 1
        # in case the any of the proposals need information about the history
        # of the chain:
        self.proposal_dist.update(self)
        return self



# Copyright (C) 2019  Collin Capano
# This program is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 3 of the License, or (at your
# option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
# Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

"""EPSIE is a toolkit for doing MCMCs using embarrasingly parallel Markov
chains.
"""

from __future__ import absolute_import

# get the version number
from ._version import __version__

import os
import sys
import pickle
from io import BytesIO
import numpy
from randomgen import PCG64

#
# =============================================================================
#
#                     Random number generation utilities
#
# =============================================================================
#

# The bit generator used for all random number generation.
# Users may change this, but it has to be something recognized by randomgen
# and have the ability to accept streams; i.e., it must have calling structure
# class(seed, stream, mode), with default stream set to None.
BIT_GENERATOR = PCG64


def create_seed(seed=None):
    """Creates a seed for a random number generator.

    Parameters
    ----------
    seed : int, optional
        If a seed is given, will just return it. Default is None, in which case
        a seed is created.

    Returns
    -------
    int :
        A seed to use.
    """
    if seed is None:
        # use os.urandom to get a string of random 4 bytes
        bseed = os.urandom(4)
        # convert to int
        if sys.version_info < (3,):
            # py27
            seed = sum([ord(c) << (i * 8) for i, c in enumerate(bseed[::-1])])
        else:
            # Py3XX
            seed = int.from_bytes(bseed, byteorder='big')
    return seed


def create_bit_generator(seed=None, stream=0):
    """Creates a an instance of :py:class:`epsie.BIT_GENERATOR`.

    Parameters
    ----------
    seed : int, optional
        The seed to use. If seed is None (the default), will create a seed
        using ``create_seed``.
    stream : int, optional
        The stream to create the bit generator for. This allows multiple
        generators to exist with the same seed, but that produce different sets
        of random numbers. Default is 0.
    """
    if seed is None:
        seed = create_seed(seed)
    try:
        return BIT_GENERATOR(seed, stream, mode="sequence")
    except TypeError as e:
        # eaerlier versions of randomgen (used for py27) did not support a
        # mode argument
        import randomgen
        if float('.'.join(randomgen.__version__.split('.')[:2])) < 1.17:
            return BIT_GENERATOR(seed, stream)
        else:
            raise e


def create_bit_generators(seed, ngenerators):
    """Creates a collection of random bit generators.

    The bit generators are different streams with the same seed. They are all
    statistically independent of each other, while still being reproducable.
    """
    return [BIT_GENERATOR(seed, ii, mode="sequence")
            for ii in range(ngenerators)]


#
# =============================================================================
#
#                          Array utilities
#
# =============================================================================
#


def array2dict(array):
    """Converts a structured array into a dictionary."""
    fields = array.dtype.names  # raises an AttributeError if array is None
    if fields is None:
        # not a structred array, just return
        return array
    return {f: _getatomic(array[f]) for f in fields}


def _getatomic(val):
    """Checks if a given value is numpy scalar. If so, it returns
    the value as its native python type.
    """
    if isinstance(val, numpy.ndarray) and val.size == 1 and val.ndim == 0:
        val = val.item(0)
    return val


#
# =============================================================================
#
#                          Parallel tempering utilities
#
# =============================================================================
#


def make_betas_ladder(ntemps, maxtemp):
    """Makes a log spaced ladder of betas."""
    minbeta = 1./maxtemp
    return numpy.geomspace(minbeta, 1., num=ntemps)


#
# =============================================================================
#
#                          Checkpointing utilities
#
# =============================================================================
#


def dump_state(state, fp, path=None, dsetname='sampler_state', protocol=None):
    """Dumps the given state to an hdf5 file handler.

    The state is stored as a raw binary array to ``{path}/{dsetname}`` in the
    given hdf5 file handler. If a dataset with the same name and path is
    already in the file, the dataset will be resized and overwritten with the
    new state data.

    Parameters
    ----------
    state : any picklable object
        The sampler state to dump to file. Can be the object returned by
        any of the samplers' `.state` attribute (a dictionary of dictionaries),
        or any picklable object.
    fp : h5py.File
        An open hdf5 file handler. Must have write capability enabled.
    path : str, optional
        The path (group name) to store the state dataset to. Default (None)
        will result in the array being stored to the top level.
    dsetname : str, optional
        The name of the dataset to store the binary array to. Default is
        ``sampler_state``.
    protocol : int, optional
        The protocol version to use for pickling. See the :py:mod:`pickle`
        module for more details.
    """
    memfp = BytesIO()
    pickle.dump(state, memfp, protocol=protocol)
    dump_pickle_to_hdf(memfp, fp, path=path, dsetname=dsetname)


def dump_pickle_to_hdf(memfp, fp, path=None, dsetname='sampler_state'):
    """Dumps pickled data to an hdf5 file object.

    Parameters
    ----------
    memfp : file object
        Bytes stream of pickled data.
    fp : h5py.File
        An open hdf5 file handler. Must have write capability enabled.
    path : str, optional
        The path (group name) to store the state dataset to. Default (None)
        will result in the array being stored to the top level.
    dsetname : str, optional
        The name of the dataset to store the binary array to. Default is
        ``sampler_state``.
    """
    memfp.seek(0)
    bdata = numpy.frombuffer(memfp.read(), dtype='S1')
    if path is not None:
        fp = fp[path]
    if dsetname not in fp:
        fp.create_dataset(dsetname, shape=bdata.shape, maxshape=(None,),
                          dtype=bdata.dtype)
    elif bdata.size != fp[dsetname].shape[0]:
        fp[dsetname].resize((bdata.size,))
    fp[dsetname][:] = bdata


def load_state(fp, path=None, dsetname='sampler_state'):
    """Loads a sampler state from the given hdf5 file object.

    The sampler state is expected to be stored as a raw bytes array which can
    be loaded by pickle.

    Parameters
    ----------
    fp : h5py.File
        An open hdf5 file handler.
    path : str, optional
        The path (group name) that the state data is stored to. Default (None)
        is to read from the top level.
    dsetname : str, optional
        The name of the dataset that the state data is stored to. Default is
        ``sampler_state``.
    """
    if path is not None:
        fp = fp[path]
    bdata = fp[dsetname][()].tobytes()
    return pickle.load(BytesIO(bdata))



# Copyright (C) 2019  Collin Capano
# This program is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 3 of the License, or (at your
# option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
# Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

"""Base class for Markov chains."""

from __future__ import absolute_import

from abc import (ABCMeta, abstractmethod, abstractproperty)
from six import add_metaclass
import numpy


@add_metaclass(ABCMeta)
class BaseChain(object):
    """Abstract base class for Markov chains.

    Provides standard functions for Chain and ParallelTemperedChain.

    Attributes
    ----------
    parameters
    iteration
    lastclear
    scratchlen
    positions
    stats
    acceptance
    blobs
    start_position
    stats0
    blob0
    current_position
    current_stats
    current_blob
    bit_generator
    random_state
    state
    hasblobs
    chain_id : int
        Integer identifying the chain. Default is 0.
    """
    chain_id = 0

    @property
    def parameters(self):
        """The sampled parameters.

        The parameters are stored as a sorted tuple.
        """
        return self._parameters

    @parameters.setter
    def parameters(self, parameters):
        if not isinstance(parameters, (list, tuple, numpy.ndarray)):
            parameters = [parameters]
        self._parameters = tuple(sorted(parameters))

    def __len__(self):
        return self.iteration - self.lastclear

    # Py3XX: uncomment the next two lines
    # @property
    # @abstractmethod
    @abstractproperty  # Py3XX: delete line
    def hasblobs(self):
        """Whether the model returns blobs."""
        pass

    # Py3XX: uncomment the next two lines
    # @property
    # @abstractmethod
    @abstractproperty  # Py3XX: delete line
    def iteration(self):
        """The number of times the chain has been stepped."""
        pass

    # Py3XX: uncomment the next two lines
    # @property
    # @abstractmethod
    @abstractproperty  # Py3XX: delete line
    def lastclear(self):
        """Returns the iteration of the last time the chain memory was cleared.
        """
        pass

    # Py3XX: uncomment the next two lines
    # @property
    # @abstractmethod
    @abstractproperty  # Py3XX: delete line
    def scratchlen(self):
        """The length of the scratch space used."""
        pass

    # Py3XX: uncomment the next two lines
    # @property
    # @abstractmethod
    @abstractproperty  # Py3XX: delete line
    def start_position(self):
        """Dictionary mapping parameters to their start position."""
        pass

    # Py3XX: uncomment the next two lines
    # @property
    # @abstractmethod
    @abstractproperty  # Py3XX: delete line
    def stats0(self):
        """Dictionary of the log likelihood and prior at the start position."""
        pass

    # Py3XX: uncomment the next two lines
    # @property
    # @abstractmethod
    @abstractproperty  # Py3XX: delete line
    def blob0(self):
        """Dictionary of the blob data at the start position."""
        pass

    # Py3XX: uncomment the next two lines
    # @property
    # @abstractmethod
    @abstractproperty  # Py3XX: delete line
    def positions(self):
        """The history of all of the positions, as a structred array."""
        pass

    # Py3XX: uncomment the next two lines
    # @property
    # @abstractmethod
    @abstractproperty  # Py3XX: delete line
    def stats(self):
        """The log likelihoods and log priors of the positions, as a structred
        array.
        """
        pass

    # Py3XX: uncomment the next two lines
    # @property
    # @abstractmethod
    @abstractproperty  # Py3XX: delete line
    def acceptance(self):
        """The history of all of acceptance ratios and accepted booleans, as
        a structred array.
        """
        pass

    # Py3XX: uncomment the next two lines
    # @property
    # @abstractmethod
    @abstractproperty  # Py3XX: delete line
    def blobs(self):
        """The history of all of the blob data, as a structred array.

        If the model does not return blobs, this is just ``None``.
        """
        pass

    # Py3XX: uncomment the next two lines
    # @property
    # @abstractmethod
    @abstractproperty  # Py3XX: delete line
    def current_position(self):
        """Dictionary of the current position of the chain."""
        pass

    # Py3XX: uncomment the next two lines
    # @property
    # @abstractmethod
    @abstractproperty  # Py3XX: delete line
    def current_stats(self):
        """Dictionary giving the log likelihood and log prior of the current
        position.
        """
        pass

    # Py3XX: uncomment the next two lines
    # @property
    # @abstractmethod
    @abstractproperty  # Py3XX: delete line
    def current_blob(self):
        """Dictionary of the blob data of the current position.

        If the model does not return blobs, just returns ``None``.
        """
        pass

    @abstractmethod
    def clear(self):
        """Clears memory of the current chain, and sets start position to the
        current position.
        """
        pass

    @abstractmethod
    def __getitem__(self, index):
        """Returns all of the chain data at the requested index."""
        pass

    # Py3XX: uncomment the next two lines
    # @property
    # @abstractmethod
    @abstractproperty  # Py3XX: delete line
    def bit_generator(self):
        """The random bit generator being used."""
        pass

    # Py3XX: uncomment the next two lines
    # @property
    # @abstractmethod
    @abstractproperty  # Py3XX: delete line
    def random_generator(self):
        """Returns the random number generator."""
        pass

    # Py3XX: uncomment the next two lines
    # @property
    # @abstractmethod
    @abstractproperty  # Py3XX: delete line
    def random_state(self):
        """The current state of the random bit generator."""
        pass

    # Py3XX: uncomment the next two lines
    # @property
    # @abstractmethod
    @abstractproperty  # Py3XX: delete line
    def state(self):
        """Returns the current state of the chain.

        The state consists of everything needed such that setting a chain's
        state using another's state will result in identical results.
        """
        pass

    @abstractmethod
    def set_state(self, state):
        """Sets the state of the chain using the given dict.

        .. warning::
           Running this will clear the chain's current memory, and replace its
           current position with what is saved in the state.

        Parameters
        ----------
        state : dict
            Dictionary of state values.
        """
        pass

    @abstractmethod
    def step(self):
        """Evolves the chain by a single step."""
        pass



# Copyright (C) 2019  Collin Capano
# This program is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 3 of the License, or (at your
# option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
# Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

"""Markov chains with parallel tempering."""

from __future__ import absolute_import

import logging
import numpy
import copy

import epsie

from .base import BaseChain
from .chain import Chain
from .chaindata import (ChainData, detect_dtypes)


class ParallelTemperedChain(BaseChain):
    """A collection of parallel tempered Markov chains.

    Parameters
    ----------
    parameters : list or tuple
        List of the parameter names to sample.
    model : object
        Any object that can be called with keyword arguments that map parameter
        names to values. When called, the object must a tuple of ``logl, logp``
        where ``logp`` is the log prior and ``logl`` is the log likelihood at
        the given point. The model may optionally return a dictionary in
        addition that maps strings to any arbitrary data.
    proposals : list of epsie.proposals instances
        List of proposal classes to use for the parameters. There must be one
        and only one proposal for every parameter. A single proposal may cover
        multiple parameters. Proposals must be instances of classes that
        inherit from :py:class:`epsie.proposals.BaseProposal`.
    betas : array of floats, optional
        Array of inverse temperatures. Each beta must be in range 0 (= infinite
        temperature; i.e., only sample the prior) <= beta <= 1 (= coldest
        temperate; i.e., sample the standard posterior). Default is a single
        beta = 1.
    swap_interval : int, optional
        For a parallel tempered chain, how often to calculate temperature
        swaps. Default is 1 (= swap on every iteration).
    bit_generator : :py:class:`epsie.BIT_GENERATOR` instance, optional
        Use the given random bit generator for generating random variates. If
        an int or None is provided, a generator will be created instead using
        ``bit_generator`` as a seed.
    chain_id : int, optional
        An interger identifying which chain this is. Default is 0.

    Attributes
    ----------
    iteration
    lastclear
    scratchlen
    positions
    stats
    acceptance
    blobs
    start_position
    stats0
    blob0
    current_position
    current_stats
    current_blob
    bit_generator
    random_state
    state
    hasblobs
    chain_id : int or None
        Integer identifying the chain.
    """
    def __init__(self, parameters, model, proposals, betas=1., swap_interval=1,
                 bit_generator=None, chain_id=0):
        self.parameters = parameters
        self.model = model
        # store the temp
        self._betas = None
        self.betas = betas
        self.swap_interval = swap_interval
        self._temperature_acceptance = None
        self._temperature_swaps = None
        if self.ntemps > 1:
            # we pass ntemps=ntemps-1 here because there will be ntemps-1
            # acceptance ratios for ntemp levels
            self._temperature_acceptance = ChainData(
                ['acceptance_ratio'], dtypes={'acceptance_ratio': float},
                ntemps=self.ntemps-1)
            self._temperature_swaps = ChainData(
                ['swap_index'], dtypes={'swap_index': int},
                ntemps=self.ntemps)
        self.chain_id = chain_id
        # make sure all parallel tempered chains use the same bit_generator
        self._bit_generator = None
        self._random_generator = None
        self.bit_generator = bit_generator
        # create a chain for each temperature
        self.chains = [
            Chain(parameters, model,
                  [copy.deepcopy(p) for p in proposals],
                  bit_generator=self.bit_generator, chain_id=chain_id,
                  beta=beta)
            for beta in self.betas]

    @property
    def bit_generator(self):
        """The random bit generator being used."""
        return self._bit_generator

    @bit_generator.setter
    def bit_generator(self, bit_generator=None):
        """Sets the random bit generator

        Parameters
        ----------
        bit_generator : :py:class:`epsie.BIT_GENERATOR` instance, optional
            Use the given random bit generator for generating random variates.
            If an int or None is provided, a generator will be created instead
            using ``bit_generator`` as a seed.
        """
        if bit_generator is None:
            bit_generator = epsie.create_bit_generator(None,
                                                       stream=self.chain_id)
        self._bit_generator = bit_generator

    @property
    def random_generator(self):
        """Returns the random number generator."""
        return self.chains[0].random_generator

    @property
    def random_state(self):
        """The current state of the random bit generator."""
        return self.bit_generator.state

    @random_state.setter
    def random_state(self, state):
        """Sets the state of bit_generator.

        Parameters
        ----------
        state : dict
            Dictionary giving the state to set.
        """
        self.bit_generator.state = state

    @property
    def state(self):
        """Returns the current state of the chain.

        The state consists of everything needed such that setting a chain's
        state using another's state will result in identical results.

        Returns
        -------
        dict :
            Dictionary of ``tk -> chains[tk].state``, where ``tk`` is the
            index of each temperature chain.
        """
        return {tk: chain.state for tk, chain in enumerate(self.chains)}

    def set_state(self, state):
        """Sets the state of the chain using the given dict.

        .. warning::
           Running this will clear the chain's current memory, and replace its
           current position with what is saved in the state.

        Parameters
        ----------
        state : dict
            Dictionary of ``tk -> dict`` mapping indices of the temperature
            chains to the state they should be set to.
        """
        for tk in state:
            self.chains[tk].set_state(state[tk])

    @property
    def hasblobs(self):
        """Whether the model returns blobs."""
        return self.chains[0].hasblobs

    @property
    def iteration(self):
        """The number of times the chain has been stepped."""
        return self.chains[0].iteration

    @property
    def lastclear(self):
        """Returns the iteration of the last time the chain memory was cleared.
        """
        return self.chains[0].lastclear

    @property
    def scratchlen(self):
        """The length of the scratch space used."""
        return self.chains[0].scratchlen

    @scratchlen.setter
    def scratchlen(self, n):
        """Set the scratch length to the given value.

        This will immediately increase the scratch length to ``n``. If the
        chain is already longer than ``n``, this will have no immediate impact.
        However, the next time ``clear`` is called, the scratch length will
        be reset to ``n``.

        Parameters
        ----------
        n : int
            The length to set.
        """
        for chain in self.chains:
            chain.scratchlen = n
        if self.ntemps > 1:
            try:
                self._temperature_swaps.set_len(n//self.swap_interval)
                self._temperature_acceptance.set_len(n//self.swap_interval)
            except ValueError:
                pass

    @property
    def betas(self):
        """Returns the betas (=1 / temperatures) used by the chain."""
        return self._betas

    @betas.setter
    def betas(self, betas):
        """Checks that the betas are in the allowed range before setting."""
        if isinstance(betas, (float, int)):
            # only single temperature; turn into list so things below won't
            # break
            betas = [betas]
        if not isinstance(betas, numpy.ndarray):
            # betas is probably a list or tuple; convert to array so we can use
            # numpy functions
            betas = numpy.array(betas)
        if not (betas == 1.).any():
            logging.warning("No betas = 1 found. This means that the normal "
                            "posterior (i.e., likelihood * prior) will not be "
                            "sampled by the chain.")
        # check that all betas are in [0, 1]
        if not ((0 <= betas) & (betas <= 1)).all():
            raise ValueError("all betas must be in range [0, 1]")
        # sort from coldest to hottest and store
        self._betas = numpy.sort(betas)[::-1]  # note: this copies the betas

    @property
    def temperatures(self):
        """Returns the temperatures (= 1 / betas) used by the chain."""
        return 1./self._betas

    @property
    def ntemps(self):
        """Returns the number of temperatures used by the chain."""
        return len(self.betas)

    def _concatenate_dicts(self, attr):
        """Concatenates dictionary attributes over all of the temperatures.

        Parameters
        ----------
        attr : str
            The name of the attribute to get from the chains. The attribute
            is assumed to return a dictionary.
        """
        # we'll create a chain data instance to stack the dictionaries
        d = getattr(self.chains[0], attr)
        out = ChainData(list(d.keys()), dtypes=detect_dtypes(d))
        out.extend(self.ntemps)
        for tk, chain in enumerate(self.chains):
            out[tk] = getattr(chain, attr)
        return out.asdict()

    def _concatenate_arrays(self, attr, item=None):
        """Concatenates array attributes over all of the temperatures.

        Returned array has shape ``[ntemps x] niterations``.
        """
        if item is None:
            arrs = list(map(lambda x: getattr(x, attr), self.chains))
        else:
            arrs = list(map(lambda x: getattr(x, attr)[item], self.chains))
        return numpy.stack(arrs)

    @property
    def start_position(self):
        """Dictionary mapping parameters to their start position.

        If the start position hasn't been set, raises a ``ValueError``.
        """
        return self._concatenate_dicts('start_position')

    @start_position.setter
    def start_position(self, position):
        """Sets the starting position.

        This also evaulates the log likelihood and log prior at the starting
        position, as well as determine if the model returns blob data.

        Parameters
        ----------
        position : dict
            Dictionary mapping parameters to values. Values
            should be numpy arrays with length = ntemps.
        """
        self._start = position.copy()
        for tk, chain in enumerate(self.chains):
            posk = {param: position[param][tk] for param in position}
            chain.start_position = posk

    @property
    def stats0(self):
        """Dictionary of the log likelihood and log prior at the start
        position.

        The values of the returned dictionary are arrays of length ``ntemps``,
        ordered by increasing temperature.

        Raises a ``ValueError`` if the start position has not been set yet.
        """
        return self._concatenate_dicts('stats0')

    @property
    def blob0(self):
        """The blob data of the starting position, as a dictionary.

        If ``hasblobs`` is False, just returns None. Otherwise,  the values of
        the returned dictionary are arrays of length ``ntemps``, ordered by
        increasing temperature.

        Raises a ``ValueError`` if ``set_start`` has not been run yet.
        """
        if self.hasblobs:
            blob = self._concatenate_dicts('blob0')
        else:
            blob = None
        return blob

    @property
    def positions(self):
        """The history of all of the positions, as a structred array.

        If ``ntemps > 1``, the returned array has shape
        ``ntemps x niterations``. Otherwise, the returned array has shape
        ``niterations``.
        """
        return self._concatenate_arrays('positions')

    @property
    def stats(self):
        """The history of all of the stats, as a structred array.

        If ``ntemps > 1``, the returned array has shape
        ``ntemps x niterations``. Otherwise, the returned array has shape
        ``niterations``.
        """
        return self._concatenate_arrays('stats')

    @property
    def acceptance(self):
        """The history of all of acceptance ratios and accepted booleans,
        as a structred array.

        If ``ntemps > 1``, the returned array has shape
        ``ntemps x niterations``. Otherwise, the returned array has shape
        ``niterations``.
        """
        return self._concatenate_arrays('acceptance')

    @property
    def temperature_acceptance(self):
        """The history of the acceptance ratios between temperatures.

        The returned array has shape
        ``ntemps-1 x (niterations/swap_interval)`` if ``ntemps > 1``.
        Otherwise, returns None.

        .. note::
           This does not return a structured array, since there is only
           one field.
        """
        if self._temperature_acceptance is None:
            return None
        out = self._temperature_acceptance[:(len(self)//self.swap_interval)]
        return out['acceptance_ratio'].T

    @property
    def temperature_swaps(self):
        """The history of all of the temperature swaps.

        The returned array has shape ``ntemps x (niterations/swap_interval)``
        if ``ntemps > 1``. Otherwise, returns None.

        .. note::
           This does not return a structured array, since there is only
           one field.
        """
        if self._temperature_swaps is None:
            return None
        out = self._temperature_swaps[:(len(self)//self.swap_interval)]
        return out['swap_index'].T

    @property
    def blobs(self):
        """The history of all of the blob data, as a structured array.

        If the model does not return blobs, this is just ``None``.

        If ``ntemps > 1``, the returned array has shape
        ``ntemps x niterations``. Otherwise, the returned array has shape
        ``niterations``.
        """
        if self.hasblobs:
            blobs = self._concatenate_arrays('blobs')
        else:
            blobs = None
        return blobs

    @property
    def current_position(self):
        """Dictionary of the current position of the chain."""
        return self._concatenate_dicts('current_position')

    @property
    def current_stats(self):
        """Dictionary giving the log likelihood and log prior of the current
        position.
        """
        return self._concatenate_dicts('current_stats')

    @property
    def current_blob(self):
        """Dictionary of the blob data of the current position.

        If the model does not return blobs, just returns ``None``.
        """
        if not self.hasblobs:
            blob = None
        else:
            blob = self._concatenate_dicts('current_blob')
        return blob

    def clear(self):
        """Clears memory of the current chain, and sets start position to the
        current position.

        New scratch space will be created with length equal to ``scratch_len``.
        """
        for chain in self.chains:
            chain.clear()
        # clear temperature swaps
        if self.ntemps > 1:
            tlen = self.scratchlen//self.swap_interval
            self._temperature_acceptance.clear(tlen)
            self._temperature_swaps.clear(tlen)

    def __getitem__(self, index):
        """Returns all of the chain data at the requested index."""
        out = {'positions': self.positions[index],
               'stats': self.stats[index],
               'acceptance': self.acceptance[index]
               }
        if self.ntemps > 1:
            out['temperature_swaps'] = \
                self.temperature_swaps[index//self.swap_interval]
            out['temperature_acceptance'] = \
                self.temperature_acceptance[index//self.swap_interval]
        if self._hasblobs:
            out['blobs'] = self.blobs[index]
        return out

    def step(self):
        """Evolves all of the temperatures by one iteration.
        """
        for chain in self.chains:
            chain.step()
        # do temperature swaps
        if self.ntemps > 1 and self.iteration % self.swap_interval == 0:
            self.swap_temperatures()

    def swap_temperatures(self):
        """Computes acceptance ratio between temperatures and swaps them.

        The positions, stats, and (if they exist) blobs are swapped. The
        acceptance is not swapped, however.
        """
        # get values of all temps at current step
        stats = self.current_stats
        # we'll create an array of indices to keep track of where things
        # will go after all of the swaps have been done
        swap_index = numpy.arange(self.ntemps, dtype=int)
        # swaps are determined by finding acceptance ratio:
        # A_jk = min( (L_j/L_k)^(beta_k - beta_j), 1)
        # We start with the hottest chain:
        loglk = stats['logl'][-1]
        # to store acceptance ratios and swaps
        ars = numpy.zeros(self.ntemps-1)
        # since stored coldest to hottest, the folling = beta_k - beta_j
        dbetas = numpy.diff(self.betas)
        # now cycle down through the temps, comparing the current one
        # to the one below it
        for tk in range(self.ntemps-1, 0, -1):
            swk = swap_index[tk]
            tj = tk - 1
            loglj = stats['logl'][tj]
            swj = swap_index[tj]
            ar = numpy.exp(dbetas[tj]*(loglj - loglk))
            u = self.random_generator.uniform()
            swap = u <= ar
            if swap:
                # move the colder index into the current slot...
                swap_index[tk] = swj
                # ...and the hotter index into the colder slot
                swap_index[tj] = swk
                # we won't change loglk so that it will get compared to
                # next coldest temperature on the next loop
            else:
                # don't swap, so drop the current loglk here, and pick up
                # the colder logl to compare on the next loop
                loglk = loglj
            # store the acceptance ratio
            ars[tj] = ar
        # now do the swaps and store
        new_positions = [self.chains[swk].current_position
                         for swk in swap_index]
        new_stats = [self.chains[swk].current_stats
                     for swk in swap_index]
        if self.hasblobs:
            new_blobs = [self.chains[swk].current_blob
                         for swk in swap_index]
        # note: we're not swapping the acceptance ratios
        ii = self.iteration - self.lastclear - 1
        for (tk, chain) in enumerate(self.chains):
            chain._positions[ii] = new_positions[tk]
            chain._stats[ii] = new_stats[tk]
            if self.hasblobs:
                chain._blobs[ii] = new_blobs[tk]
        self._temperature_acceptance[ii//self.swap_interval] = {
            'acceptance_ratio': ars}
        self._temperature_swaps[ii//self.swap_interval] = {
            'swap_index': swap_index}



# Copyright (C) 2019  Collin Capano
# This program is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 3 of the License, or (at your
# option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
# Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

"""Classes for parallel tempered Markov chains."""

from __future__ import absolute_import

import numpy
import copy

from epsie import create_bit_generator
from epsie.chain import ParallelTemperedChain
from epsie.chain.chaindata import (ChainData, detect_dtypes)

from .base import BaseSampler


class ParallelTemperedSampler(BaseSampler):
    """Evolves a collection of parallel tempered Markov chains.

    Parameters
    ----------
    parameters : tuple or list
        Names of the parameters to sample.
    model : object
        Model object.
    nchains : int
        The number of chains to create per temperature. Must be greater than
        zero.
    betas : numpy.ndarray of floats
        The betas (= 1 / temperatures) to use. All betas must be between [0,1].
        Must provide at least one. If one is not included in the betas, a
        warning will be printed.
    swap_interval : int, optional
        How often to calculate temperature swaps. Default is 1 (= swap on every
        iteration).
    proposals : dict, optional
        Dictionary mapping parameter names to proposal classes. Any parameters
        not listed will use the ``default_propsal``.
    default_proposal : an epsie.Proposal class, optional
        The default proposal to use for parameters not in ``proposals``.
        Default is :py:class:`epsie.proposals.Normal`.
    default_proposal_args : dict, optional
        Dictionary of arguments to pass to the default proposal.
    seed : int, optional
        Seed for the random number generator. If None provided, will create
        one.
    pool : Pool object, optional
        Specify a process pool to use for parallelization. Default is to use a
        single core.
    """
    def __init__(self, parameters, model, nchains, betas, swap_interval=1,
                 proposals=None,
                 default_proposal=None, default_proposal_args=None, seed=None,
                 pool=None):
        self.parameters = parameters
        self.model = model
        self.set_proposals(proposals, default_proposal, default_proposal_args)
        self.seed = seed
        self.set_map(pool)
        if isinstance(betas, (float, int)):
            # only single temperature; turn into list so things below won't
            # break
            betas = [betas]
        if not isinstance(betas, numpy.ndarray):
            # betas is probably a list or tuple; convert to array so we can use
            # numpy functions
            betas = numpy.array(betas)
        self.create_chains(nchains, betas, swap_interval)

    def create_chains(self, nchains, betas, swap_interval=1):
        """Creates a list of :py:class:`chain.ParallelTemperedChain`.

        Parameters
        ----------
        nchains : int
            The number of parallel tempered chains to create.
        betas : array
            Array of inverse temperatures to use for each parallel tempered
            chain.
        swap_interval : int, optional
            How often to calculate temperature swaps. Default is 1 (= swap on
            every iteration).
        """
        if nchains < 1:
            raise ValueError("nchains must be >= 1")
        self._chains = [ParallelTemperedChain(
            self.parameters, self.model,
            [copy.deepcopy(p) for p in self.proposals.values()],
            betas=betas, swap_interval=swap_interval,
            bit_generator=create_bit_generator(self.seed, stream=cid),
            chain_id=cid)
            for cid in range(nchains)]

    @property
    def betas(self):
        """The betas used."""
        return self.chains[0].betas

    @property
    def ntemps(self):
        """Returns the number of temperatures used."""
        return len(self.betas)

    @property
    def temperatures(self):
        """The temperatures used."""
        return 1./betas

    @property
    def swap_interval(self):
        """Returns the swap interval used for parallel tempering."""
        return self.chains[0].swap_interval

    @swap_interval.setter
    def swap_interval(self, interval):
        """Sets the swap interval to use."""
        for c in self.chains:
            c.swap_interval = interval

    def _concatenate_dicts(self, attr):
        """Concatenates dictionary attributes over all of the chains.

        This is a convenience function used by properties such as
        ``current_positions`` to gather all of the dictionary attributes from
        the chains.

        Parameters
        ----------
        attr : str
            The name of the attribute to get from the chains. The attribute
            is assumed to return a dictionary.

        Returns
        -------
        dict :
            Dictionary mapping parameters to arrays. The arrays have shape
            ``ntemps x nchains``.
        """
        # we'll create a chain data instance to stack the dictionaries
        d = getattr(self.chains[0], attr)
        out = ChainData(list(d.keys()), dtypes=detect_dtypes(d),
                        ntemps=self.ntemps)
        out.extend(self.nchains)
        for ii, chain in enumerate(self.chains):
            out[ii] = getattr(chain, attr)
        return out.asdict()

    def _concatenate_arrays(self, attr, item=None):
        """Concatenates the given attribute over all of the chains.

        This is a convenience function used by properties such as ``positions``
        to gather all of the array attributes from the chains.

        Parameters
        ----------
        attr : str
            The name of the attribute to get from the chains. The attribute
            is assumed to return a (structred) array.
        item : str or array index, optional
            Get a particular item from the (structred) array from each chain
            before concatenating.

        Returns
        -------
        array :
            The returned array has shape ``ntemps x nchains x niterations``.
        """
        if item is None:
            arrs = list(map(lambda x: getattr(x, attr), self.chains))
        else:
            arrs = list(map(lambda x: getattr(x, attr)[item], self.chains))
        return numpy.stack(arrs, axis=1)

    @property
    def temperature_acceptance(self):
        """The history of the temperature acceptance from all of the chains.

        If ntemps is 1, just returns None.

        .. note::
           This does not return a structured array, since there is only
           one field.
        """
        if self.ntemps == 1:
            return None
        return self._concatenate_arrays('temperature_acceptance')

    @property
    def temperature_swaps(self):
        """The history of all the temperature swaps from all of the chains.
        
        If ntemps is 1, just returns None.

        .. note::
           This does not return a structured array, since there is only
           one field.
        """
        if self.ntemps == 1:
            return None
        return self._concatenate_arrays('temperature_swaps')



# Copyright (C) 2019  Collin Capano
# This program is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 3 of the License, or (at your
# option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
# Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

# Py3XX: delete abstractproperty
from abc import ABCMeta, abstractmethod, abstractproperty
from six import add_metaclass

import numpy
try:
    from randomgen import RandomGenerator
except ImportError:
    from randomgen import Generator as RandomGenerator
from scipy import stats

import epsie


@add_metaclass(ABCMeta)
class BaseProposal(object):
    """Abstract base class for all proposal classes.

    All proposals must inherit from this class, as it handles random number
    generation, and lays out all of the functions/attributes that a sampler
    will try to access.

    .. warning ::
        All proposals must use the random number generator provided by this
        class (see the ``random_generator`` attribute) for creating random
        numbers. **Do not attempt to use scipy/numpy's random number
        generator.** Use of any other generator may result in chains not being
        independent of each other when run in a parallel environment.

    In addition to the abstract methods/properties, all samplers must set
    a ``parameters`` attribute. This is a list of the names of the parameters
    the proposal produces jumps for.

    Attributes
    ----------
    bit_generator
    random_generator
    random_state
    parameters
    symmetric
    state
    """
    name = None
    _parameters = None

    @property
    def bit_generator(self):
        """The random bit generator instance being used.

        A bit generator will be created if it doesn't exist yet.
        """
        try:
            return self._bit_generator
        except AttributeError:
            self._bit_generator = epsie.create_bit_generator()
            return self._bit_generator

    @bit_generator.setter
    def bit_generator(self, bit_generator):
        """Sets the random bit generator.

        Parameters
        ----------
        bit_generator : :py:class:`epsie.BIT_GENERATOR`, int, or None
            Either the bit generator to use or an integer/None. If the latter,
            a generator will be created by passing ``bit_generator`` as the
            ``seed`` argument to :py:func:`epsie.create_bit_generator`.
        """
        if not isinstance(bit_generator, epsie.BIT_GENERATOR):
            bit_generator = epsie.create_bit_generator(bit_generator)
        self._bit_generator = bit_generator

    @property
    def random_generator(self):
        """The random number generator.

        This is an instance of :py:class:`randgen.RandomGenerator` that is
        derived from the bit generator. It provides methods to create random
        draws from various distributions.
        """
        return RandomGenerator(self.bit_generator)

    @property
    def random_state(self):
        """The current state of the random bit generator.
        """
        return self.bit_generator.state

    @random_state.setter
    def random_state(self, state):
        """Sets the state of bit_generator.
        
        Parameters
        ----------
        state : dict
            Dictionary giving the state to set.
        """
        self.bit_generator.state = state

    @property
    def parameters(self):
        """Sorted tuple of the parameters that proposals are produced for."""
        if self._parameters is None:
            raise AttributeError("no parameters set")
        return self._parameters

    @parameters.setter
    def parameters(self, parameters):
        """Sets the parameters.

        The parameters are stored as a sorted tuple.

        Parameters
        ----------
        parameters : (list of) str
            The names of the parameters. This may either be a list of strings,
            or (for a single parameter), a string.
        """
        if not isinstance(parameters, (list, tuple, numpy.ndarray)):
            parameters = [parameters]
        self._parameters = tuple(sorted(parameters))

    # Py3XX: uncomment the next two lines
    # @property
    # @abstractmethod
    @abstractproperty  # Py3XX: delete line
    def symmetric(self):
        """Boolean indicating whether the proposal distribution is symmetric.

        A jump proposal is symmetric if the proposal probability
        density has the property that :math:`p(x'|x) = p(x|x')`,
        where :math:`x` is the current position, :math:`x'` is the proposed
        position. In this, case, the Metropolis-Hastings ratio simplfies to
        just the ratio of posterior values at the two points.

        Note that an adaptive proposal may still be symmetric. Symmetry only
        means that the current state of the proposal satisfies the above
        condition; it does not necessarily mean that transitions between states
        be symmetric.
        """
        pass

    # Py3XX: uncomment the next two lines
    # @property
    # @abstractmethod
    @abstractproperty  # Py3XX: delete line
    def state(self):
        """Returns all information needed to produce a deterministic jump.

        The information returned by this property should be everything needed
        such that if you pass it to set_state, you will get the same proposal
        on the next call of jump.

        The information should be returned as a dictionary. At the very least,
        this should include the current state of the proposal's
        ``random_state``. For adaptive proposals, this may also include the
        buffer used to adjust the proposal distribution.
        """
        pass

    @abstractmethod
    def set_state(self, state):
        """Set all information needed to produce a deterministic jump.
        """
        pass

    @abstractmethod
    def jump(self, fromx):
        """This should provide random samples from the proposal distribution.

        Samples should be returned as a dictionary mapping parameters to
        the proposed jump.
        """
        pass

    @abstractmethod
    def logpdf(self, xi, givenx):
        """The log pdf of the proposal distribution at a point.

        Parameters
        ----------
        xi : dict
            Dictionary mapping parameter names to values to evaluate.
        givenx : dict, optional
            Dictionary mapping parameter names to values of the point from
            which ``xi`` is evaluated.
        """
        pass

    def pdf(self, xi, givenx):
        """The pdf of proposal at the given values.

        This just expoentiates ``logpdf``.

        Parameters
        ----------
        xi : dict
            Dictionary mapping parameter names to values to evaluate.
        givenx : dict, optional
            Dictionary mapping parameter names to values of the point from
            which ``xi`` is evaluated.
        """
        return numpy.exp(self.logpdf(xi, givenx))

    def update(self, chain):
        """Update the state of the proposal distribution after a jump.

        This method may optionally be implemented by a proposal. It is called
        by the Markov chains just after a jump is evaluated. It can be used by,
        e.g., adaptive jump proposals that change their state depending on
        the history of the chain.
        """
        pass



#!/usr/bin/env python

# Copyright (C) 2019 Collin Capano
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 2 of the License, or (at your
# option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
# Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

"""
setup.py file for PyCBC package
"""
import setuptools


install_requires = ["numpy",
                    "scipy>=0.16.0",
                    'randomgen<1.17; python_version < "3.0"',
                    'randomgen>=1.18; python_version >= "3.0"',
                    'six>1.10.0',
                   ]

with open("README.md", "r") as fh:
    long_description = fh.read()

# get version
with open("epsie/_version.py", "r") as fh:
    vstr = fh.read().strip()
try:
    vstr = vstr.split('=')[1].strip()
except IndexError:
    raise RuntimeError("version string in epsie._verion.py not formatted "
                       "correctly; it should be:\n"
                       "__version__ = VERSION")
# remove ' and " from version string
version = vstr.replace("'", "").replace('"', "")

setuptools.setup(
    name="epsie",
    version=version,
    author="Collin D. Capano",
    author_email="cdcapano@gmail.com",
    description="An Embarrassingly Parallel Sampler for "
                "Inference Estimation.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/cdcapano/epsie",
    install_requires = install_requires,
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python :: 3.6",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Operating System :: OS Independent",
        'Intended Audience :: Science/Research',
        'Natural Language :: English',
    ],
)
